#include <winver.h>

VS_VERSION_INFO VERSIONINFO
FILEVERSION     1,0,0,0
PRODUCTVERSION  1,0,0,0
FILEFLAGSMASK   VS_FFI_FILEFLAGSMASK
FILEFLAGS       VS_FF_PRERELEASE
FILEOS          VOS_NT_WINDOWS32
FILETYPE        VFT_APP
FILESUBTYPE     VFT2_UNKNOWN
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904B0"
        BEGIN
            VALUE "FileDescription",  "ProTracker v2.3D clone for Windows"
            VALUE "ProductName",      "ProTracker"
            VALUE "ProductVersion",   "1.0"
            VALUE "InternalName",     "ProTracker"
            VALUE "LegalCopyright",   "Public domain"
            VALUE "OriginalFilename", "protracker.exe"
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x4090, 1200
    END
END

IDI_MAIN_ICON ICON "gfx/protracker.ico"
