#include <stdint.h>
#include "../pt_palette.h"

const uint8_t arrowPaletteBMP[30] =
{
    PAL_GENBKG, PAL_GENBKG, PAL_GENBKG, PAL_GENTXT, PAL_GENBKG, PAL_GENBKG,
    PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENBKG,
    PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT,
    PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENTXT, PAL_GENBKG,
    PAL_GENBKG, PAL_GENBKG, PAL_GENBKG, PAL_GENTXT, PAL_GENBKG, PAL_GENBKG
};
