#include <stdint.h>

// Final unpack length: 3520
// Decoded length: 880 (first four bytes of buffer)
const uint8_t termTopPackedBMP[379] =
{
    0x00, 0x00, 0x03, 0x70, 0xCC, 0x08, 0x55, 0x56, 0xCC, 0x42, 0x55, 0x95, 0x55, 0x56, 0x6A, 0xCC,
    0x07, 0xAA, 0xAB, 0x6A, 0xCC, 0x41, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xCC, 0x07, 0xAA, 0xAB, 0x6A,
    0xCC, 0x41, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0x55, 0xA5, 0xAA, 0x95, 0x56, 0x95, 0x69, 0x55, 0xAB,
    0x6A, 0xCC, 0x0F, 0xAA, 0x55, 0x69, 0x55, 0xA9, 0x56, 0x95, 0x56, 0x55, 0x6A, 0x55, 0xA9, 0x56,
    0x96, 0x96, 0x55, 0x59, 0x55, 0xAA, 0xAA, 0x95, 0x56, 0x55, 0x59, 0x55, 0xA6, 0xA9, 0xA5, 0x5A,
    0x5A, 0x5A, 0x55, 0xA5, 0xCC, 0x11, 0xAA, 0xD9, 0x69, 0x6B, 0x69, 0x7D, 0x65, 0xEA, 0x97, 0xFF,
    0x5F, 0x59, 0x7D, 0x6B, 0x6A, 0xCC, 0x0F, 0xAA, 0x5F, 0x59, 0x7D, 0x65, 0xF5, 0xAD, 0x7F, 0x5F,
    0x59, 0x7D, 0x65, 0xF5, 0x97, 0x5F, 0x5F, 0xFD, 0x7D, 0x6A, 0xAA, 0xAD, 0x7F, 0x5F, 0xFD, 0x7D,
    0x65, 0xA5, 0xE9, 0x7E, 0x56, 0x5D, 0x7D, 0x65, 0xEA, 0xCC, 0x10, 0xAA, 0xDA, 0x55, 0xFB, 0x69,
    0x7A, 0xF5, 0xEA, 0x95, 0x6A, 0x55, 0x5D, 0x55, 0xFB, 0x6A, 0xCC, 0x0F, 0xAA, 0x55, 0x7D, 0x55,
    0xF5, 0xE5, 0xE9, 0x7A, 0x55, 0x7D, 0x55, 0x75, 0xEB, 0xD5, 0x7E, 0x55, 0xA9, 0x55, 0xFA, 0xAA,
    0xA9, 0x7A, 0x55, 0xA9, 0x55, 0xF5, 0x55, 0xE9, 0x7A, 0x55, 0x5D, 0x55, 0x75, 0xEA, 0xCC, 0x10,
    0xAA, 0xDA, 0x97, 0xEB, 0x69, 0x79, 0x65, 0xEA, 0x97, 0xFA, 0x5F, 0x5D, 0x75, 0xEB, 0x6A, 0xCC,
    0x0F, 0xAA, 0x5F, 0xF9, 0x75, 0xE5, 0xE5, 0xE9, 0x7A, 0x5D, 0x79, 0x7D, 0x75, 0xE5, 0x97, 0x5A,
    0x5F, 0xE9, 0x75, 0xEA, 0xAA, 0xA9, 0x7A, 0x5F, 0xE9, 0x75, 0xE5, 0xF5, 0xE9, 0x7A, 0x5D, 0x5D,
    0x7D, 0x75, 0xEA, 0xCC, 0x10, 0xAA, 0xDA, 0x55, 0xAB, 0x6A, 0x55, 0xF5, 0x55, 0x95, 0x56, 0x5E,
    0x5D, 0x79, 0x6B, 0x6A, 0xCC, 0x0F, 0xAA, 0x5E, 0xA9, 0x79, 0x69, 0x57, 0xE9, 0x7A, 0x5E, 0x59,
    0x79, 0x79, 0x57, 0xD7, 0x96, 0x55, 0x59, 0x79, 0x6A, 0xAA, 0xA9, 0x7A, 0x55, 0x59, 0x79, 0x65,
    0xE5, 0xE5, 0x5A, 0x5E, 0x5D, 0x79, 0x75, 0x55, 0xCC, 0x10, 0xAA, 0xD9, 0x7D, 0x6B, 0x6A, 0xBF,
    0xEB, 0xFF, 0xEF, 0xFF, 0xBE, 0xBE, 0xFA, 0xFB, 0x6A, 0xCC, 0x0F, 0xAA, 0xBE, 0xAA, 0xFA, 0xFA,
    0xFF, 0xAA, 0xFA, 0xBE, 0xBE, 0xFA, 0xFA, 0xFF, 0xAF, 0xAF, 0xBF, 0xFE, 0xFA, 0xFA, 0xAA, 0xAA,
    0xFA, 0xBF, 0xFE, 0xFA, 0xFB, 0xEB, 0xEB, 0xFE, 0xBE, 0xBE, 0xFA, 0xFB, 0xFF, 0xEA, 0xCC, 0x0F,
    0xAA, 0xDA, 0xFA, 0xFB, 0x6A, 0xCC, 0x07, 0xAA, 0xAB, 0x6A, 0xCC, 0x41, 0xAA, 0xDA, 0xAA, 0xAB,
    0xBF, 0xCC, 0x08, 0xFF, 0xBF, 0xCC, 0x41, 0xFF, 0xEF, 0xFF, 0xFF
};

// Final unpack length: 2552
// Decoded length: 638 (first four bytes of buffer)
const uint8_t termScrollBarPackedBMP[642] =
{
    0x00, 0x00, 0x02, 0x7E, 0x55, 0x55, 0x59, 0xAA, 0xAA, 0xB6, 0xA9, 0xAA, 0xDA, 0x95, 0xAB, 0x69,
    0x55, 0xAD, 0xA9, 0x5F, 0xB6, 0xA5, 0x7A, 0xDA, 0x95, 0xEB, 0x6A, 0xBF, 0xAD, 0xAA, 0xAA, 0xBB,
    0xFF, 0xFF, 0xD5, 0x55, 0x56, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB,
    0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA,
    0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA,
    0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA,
    0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA,
    0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD,
    0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA,
    0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA,
    0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA,
    0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A,
    0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6,
    0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB,
    0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA,
    0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA,
    0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA,
    0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA,
    0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD,
    0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA,
    0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA,
    0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA,
    0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A,
    0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6,
    0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB,
    0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA,
    0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA,
    0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA,
    0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA,
    0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD,
    0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA,
    0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA,
    0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA,
    0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A,
    0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6,
    0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB,
    0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA,
    0xB6, 0xAA, 0xAA, 0xDA, 0xAA, 0xAB, 0x6A, 0xAA, 0xAD, 0xAA, 0xAA, 0xB6, 0xAA, 0xAA, 0xDA, 0xAA,
    0xAB, 0xBF, 0xFF, 0xFD, 0x55, 0x55, 0x66, 0xAA, 0xAA, 0xDA, 0x95, 0xAB, 0x6A, 0x57, 0xAD, 0xA9,
    0x5E, 0xB6, 0x95, 0x5A, 0xDA, 0x95, 0xFB, 0x6A, 0x9F, 0xAD, 0xAA, 0xBA, 0xB6, 0xAA, 0xAA, 0xEF,
    0xFF, 0xFF
};
